/*  rm - a substitute for the standard unix 'rm' command
 *  V1.0, 29 May 1992, (C) Copyright 1992, Alessandro Bottonelli
 *
 *  La redistribuzione e l'uso in forma binaria e sorgente sono concesse
 *  a condizione che questa notifica sia mantenuta nella sua interezza e
 *  che all'autore originale sia dato il dovuto credito. Il nome dell'autore
 *  originale non potra' essere usato per promuovere prodotti derivati da
 *  questo software senza specifico consenso scritto e anticipato. Questo
 *  software e' reso disponibile "cosi come e'", senza alcuna garanzia 
 *  implicita o esplicita da parte dell'autore e del suo datore di lavoro.
 *
 *  Redistribution and use in source and binary forms are permitted provided
 *  that this notice is preserved and that due credit is given to the 
 *  original author. The name of the author may not be used to promote or
 *  endorse products derived from this software without specific prior 
 *  written permission. This software is provided 'as is' without express
 *  or implied warranty by the author or his employer. 
 *
 *  USO E REDISTRIBUZIONE DI QUESTO SOFTWARE A FINI COMMERCIALI E' ESPRESSAMENTE
 *  PROIBITO.
 *
 *  USE AND REDISTRIBUTION OF THIS SOFTWARE FOR COMMERCIAL PURPOSES IS
 *  EXPRESSELY PROHIBITED.
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>

#define  compress	"/usr/bin/compress -f"	/* Could be something else */
#define  real_rm	"/bin/rm"		/* This should be it!      */
#define  RM_DIR		"/tmp/rm"		/* Backup directory        */
#define  S_SIZE		80			/* Strings size            */
#define  USER_NAME	"LOGNAME"		/* Could be something else */

extern int errno;
char *basename();
char *getlogin();
char *getenv();

int verbose=0;		/* 0 if you like a more silent version of RM      */

main(argc, argv)
int argc;
char **argv;
{
	extern int optind;			/* for getopt()            */
	int c;					/* As Above                */
	char command[S_SIZE];			/* command for system()    */
	char login_name[S_SIZE];		/* current user name       */
	char dirname[S_SIZE];			/* name of backup dir      */
	char *cp;				/* for getenv()            */
	int oldumask; 				/* to hold umask           */
	int index;				/* Index to arguments in argc */

/*
 * Scan for options 
 * Accept them but don't do anything with them ...
 */

	while ((c = getopt(argc, argv, "rfi")) != EOF) {
	  switch (c) {
	    case 'r' : 
	    case 'f' : 
	    case 'i' : break;	
	    case '?' : exit(fprintf(stderr,"Usage: rm [-rfi] files\n"));
	  }
	}

/*
 * If no arguments after options, exit here
 */

	if ( optind == argc )
		exit(0);

/* 
 * Check that top level backup directory is available
 * try to create one if not already present
 */

	if (access(RM_DIR, R_OK | W_OK | X_OK | F_OK )) {
	  sprintf(command,"mkdir %s; chmod 777 %s", RM_DIR, RM_DIR);
	  if (system(command) != 0) {
	    printf("'%s' failed!!\n", command);
	    exit(1);
	  }
	}
/*
 * Get the user name, then determine whether a second level 
 * backup directory after the name of the user exists, if it
 * doesn't try to create one
 */
	if ((cp = getenv(USER_NAME)) == NULL)
		strcpy(login_name, getlogin());
	else
		strcpy(login_name, cp);

	sprintf(dirname, "%s/%s", RM_DIR, login_name);

	if (access(dirname, R_OK | W_OK | X_OK | F_OK )) {
	  sprintf(command,"mkdir %s; chmod 700 %s", dirname, dirname);
	  if (system(command) != 0) {
	    printf("'%s' failed!!\n", command);
	    exit(1);
	  }
	}

/* 
 * Whatever the current umask, set it to 077 octal
 * and save the current one
 */
	oldumask = umask(077);

/* 
 * Now for each file in the argument line that is
 * accessible and a regular file, back it up ...
 */
	for( index=optind; index<argc; index++)
		if (access(basename(argv[index]), R_OK | F_OK) == 0)
			if ( Regular_File( basename( argv[index])))
				save_copy_of(dirname, argv[index]);

/* 
 * When finished, set back the old umask value
 * then exec the REAL rm, if we get to the next line of
 * code after execv() something went wrong ...
 */
	(void) umask(oldumask);
	execv(real_rm, argv);
	fprintf(stderr,"rm: error %d exec'ing!\n", errno);
	exit(1);
}

/*
 * returns the basename of the file specified in string
 */

char *basename(string)
char *string;
{
	static   char *buff;

	buff = string + strlen(string);
	while (*buff != '/' && buff > string) buff--;
	return( (char *) (*buff == '/'? ++buff : buff));
}

/* save_copy_of() optimized (time-wyse) by copying 
 * file at 4KB chunks, you may want to tune this
 * on your system. 
 */

#include <sys/types.h>
#include <fcntl.h>
#define BUFSZ 4096	/* Tune this on your system for best performance */

save_copy_of(dirname, filename)
char *dirname, *filename;
{
	char backfile[S_SIZE];			/* Backup file name    */
	char buffer[BUFSZ];			/* Buffer for copy op. */
	char command[S_SIZE];			/* Command buffer      */
	int count;				/* Bytes actually read */

/*
 * Build backup name, then remove it if already present 
 */
	sprintf(backfile,"%s/%s", dirname, basename(filename));
	(void) unlink(backfile);

/*
 * Try to link it, if not possible (typically accross file systems)
 * then copy it
 */
	if ( verbose )
		fprintf(stderr,"Backing up %s ...\n", basename(backfile));

	if (link(filename, backfile) != 0) {
	  int infile, outfile;
	  int   c;
	  
	 if ((infile = open(filename, O_RDONLY)) < 0)
		exit(fprintf(stderr, "rm: can't read file '%s' to save a copy!\n", filename));

	 if ((outfile = open(backfile, O_WRONLY | O_CREAT, 0666)) < 0)
		exit(fprintf(stderr, "rm: can't write to file '%s'!\n", backfile));
/*
 *	Copy it in BUFSZ chunks. Keep count of bytes actually read
 *	from source file in order to handle the last chunk (it will
 *	probably not be exactely of BUFSZ size!)
 */
	 while( (count=read( infile, buffer, BUFSZ)) > 0 )
		if ( write( outfile, buffer, count) < count )
			exit(fprintf(stderr, "rm: write to file '%s' failed!\n", backfile));
 
	  close(infile);
	  close(outfile);
/*
 * 	Let's give it a try to compress the backup file.
 *      It may not succeed for a number of reasons (but we don't care)
 */
	sprintf(command,"%s %s", compress, backfile);
	if( verbose )
		fprintf(stderr,"Compressing %s ...\n", basename(backfile));

	if ( system(command) )
		return(fprintf(stderr,"rm: compress of %s failed!/n",backfile));
	}
}

/* Regular_File()
 * Determines whether a file in 'path' is a 
 * regular file (return 1 if so, return 0 if not)
 */

int Regular_File( path)
char *path;
{
	struct stat buf;		/* For stat()		*/	
	
	if( stat( path, &buf) != 0 )
		return(0);
	else 
		if ( buf.st_mode & _S_IFMT != _S_IFREG )
			return(0);

	return(1);
}
