/*  unrm - a companion to the non-unix-standard rm command
 *  V1.0, 29 May 1992, (C) Copyright 1992, Alessandro Bottonelli
 *
 *  La redistribuzione e l'uso in forma binaria e sorgente sono concesse
 *  a condizione che questa notifica sia mantenuta nella sua interezza e
 *  che all'autore originale sia dato il dovuto credito. Il nome dell'autore
 *  originale non potra' essere usato per promuovere prodotti derivati da
 *  questo software senza specifico consenso scritto e anticipato. Questo
 *  software e' reso disponibile "cosi come e'", senza alcuna garanzia 
 *  implicita o esplicita da parte dell'autore e del suo datore di lavoro.
 *
 *  Redistribution and use in source and binary forms are permitted provided
 *  that this notice is preserved and that due credit is given to the 
 *  original author. The name of the author may not be used to promote or
 *  endorse products derived from this software without specific prior 
 *  written permission. This software is provided 'as is' without express
 *  or implied warranty by the author or his employer. 
 *
 *  USO E REDISTRIBUZIONE DI QUESTO SOFTWARE A FINI COMMERCIALI E' ESPRESSAMENTE
 *  PROIBITO.
 *
 *  USE AND REDISTRIBUTION OF THIS SOFTWARE FOR COMMERCIAL PURPOSES IS
 *  EXPRESSELY PROHIBITED.
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#define  RM_DIR		"/tmp/rm"		/* Backup Directory         */
#define  S_SIZE		80			/* String size               */
#define uncompress	"/usr/bin/uncompress"	/* May differ on your system */
#define COMPRESS_SUFFIX ".Z"			/* As above                  */

char *getenv();
char *getlogin();
char *basename();

int verbose=0;		/* 0 if you'd like a more silent version of UNRM */
int flag_overwrite=0;	/* Overwrite existing file if TRUE               */

main(argc, argv)
int argc;
char **argv;
{
	extern int optind;		/* for getopt()                */
	int    c;			/* for getopt()                */
	char login_name[S_SIZE];	/* name of user                */
	char dirname[S_SIZE];		/* name of backup directory    */
	char *cp;			/* for getenv()                */
	int index;			/* index to file names in argc */

/*
 * Scan for options and arguments
 */
	while ((c = getopt(argc, argv, "f")) != EOF) {
	  switch (c) {
	    case 'f' : flag_overwrite++;	break;
	    case '?' : exit(fprintf(stderr,"Usage: unrm [-f] files\n"));
	  }
	}

/* 
 * If there are no file names after options, exit here
 */
	if( optind == argc )	
		exit(fprintf(stderr,"Usage: unrm [-f] files\n"));

/*
 * Checkout whether backup directory exists 
 */
	if (access(RM_DIR, F_OK | R_OK | W_OK | X_OK ))
		exit(fprintf(stderr,"Error: Directory %s doesn't exist!\n", RM_DIR));

/* 
 * Check who is trying to undelete file 
 */
	if ((cp = getenv("LOGNAME")) == NULL)
		strcpy(login_name, getlogin());
	else
		strcpy(login_name, cp);

/*
 * His/her files should be under RM_DIR/login_name directory
 * Check whether it exists ...
 */
	sprintf(dirname, "%s/%s", RM_DIR, login_name);
	if (access(dirname, F_OK | R_OK | W_OK | X_OK ))
		exit(fprintf(stderr,"Error: Directory %s doesn't exist!\n", dirname));

/* 
 * Finally. For each file in the argument line, restore it
 * then exit
 */
	for (index=optind; index<argc; index++)
		restore(dirname, argv[index]);

	exit(0);
}

/*
 * Restore routine. 
 */

#include <fcntl.h>

/* Buffer size for copy operation, tune on your system
 * for best balance between performance and memory  constraints
 */

#define BUFSZ 4096

restore(directory, filename)
char *directory, *filename;
{
	struct stat buffer;			/* for stat() routine     */
	char backfile[S_SIZE];			/* Name of backup file    */
	char backfileZ[S_SIZE];			/* name of compressed back*/
	char reply[S_SIZE];			/* User reply to questions*/
	char buf[BUFSZ];			/* Buffer for copy file   */
	char command[S_SIZE];			/* systme() command       */
	int compress_flag;  			/* flag for uncompress    */

/* 
 * Clear compress_flag and build backup file names (plain and
 * compressed)
 */
	compress_flag=0;
	sprintf(backfile,"%s/%s", directory, filename);
	sprintf(backfileZ,"%s/%s%s", directory, filename, COMPRESS_SUFFIX);

/*
 * Try to access() plain backup file, if fail try to access compressed
 * backup file, if fails again return error else set compress_flag
 */
	if (access(backfile, F_OK | R_OK) != 0) {
		compress_flag++;
		if (access(backfileZ, F_OK | R_OK) != 0) 
	  		return(fprintf(stderr,"Error: Can't find old copy of '%s'!\n", filename));
	}

/*
 * If we are dealing with a compressed backup file, stat() it, determine
 * whether it belongs to the current user, return error if it doesn't else
 * uncompress it (check whether uncompression works correctly!)
 */
	if ( compress_flag )
	{
		if (stat(backfileZ, &buffer) != 0)
	  		return(fprintf(stderr,"Error: Can't stat compressed copy of '%s'!\n", filename));
		if (buffer.st_uid != getuid())
	  		return(fprintf(stderr,"Error: File '%s' isn't yours to restore!\n", filename));
		sprintf( command, "%s %s", uncompress, backfileZ);
		
		if ( verbose )
			fprintf(stderr,"Uncompressing %s ...\n", basename(backfileZ));
		if( system(command) != 0 )
			return(fprintf(stderr,"Error: Uncompress failed!\n"));
	}

/* Now backup file should be a plain file (non compressed)
 * check again for ownership and accessability then if everything
 * is OK carry on ...
 */
	if (stat(backfile, &buffer) != 0)
	  	return(fprintf(stderr,"Error: Can't stat old copy of '%s'!\n", filename));

	if (buffer.st_uid != getuid())
	  return(fprintf(stderr,"Error: File '%s' isn't yours to restore!\n", filename));

/* Check whether a file with the same name already exists in
 * the current directory. If it does and -f option wasn't specified
 * prompt user for a decision 
 */
	if (access(filename, F_OK) == 0) {
	  if (! flag_overwrite)
	    printf( "File %s already exists in this directory!  Replace it? (y/n) ", filename);
	    gets(reply);
	    if (tolower(reply[0]) != 'y') {
	      	fprintf(stderr,"Restore of file %s cancelled\n", filename);
	        return(0);
	    }
	}

/*
 * If file exists delete it
 */
	(void) unlink(filename);

/* 
 * Try to link file before copying it 
 * if not possible (typically accross file systems) then copy it
 */
	if (link(backfile, filename) != 0) {
	  int infile, outfile;	/* File descriptors for source and dest */
	  int count;		/* Counter of bytes actually read       */
 
	 if ((infile = open(backfile, O_RDONLY)) < 0)
 	   exit(fprintf(stderr, "rm: can't read file '%s' to save a copy!\n", backfile));

	 if ((outfile = open(filename, O_WRONLY | O_CREAT, 0666)) < 0)
	   exit(fprintf(stderr, "rm: can't write to file '%s'!\n", filename));

/*       Copy in and out at BUFSZ chunks (the larger BUFSZ the better???)
 *       Write out the number of bytes actually read (deals with last
 *       chunk which will probably not be exactely BUFSZ size 
 */
	 while( (count=read( infile, buf, BUFSZ)) > 0 )
		if ( write( outfile, buf, count) < count )
			exit(fprintf(stderr, "rm: write to file '%s' failed!\n", filename));
 
	  close(infile);
	  close(outfile);
	}
/* 
 * Delete backup file (we don't need it any longer)
 * Inform user the file has been restored and return
 */
	unlink(backfile);
	if ( verbose )
		fprintf(stderr,"Restored file '%s'\n", filename);
	return(0);
}

char *basename(string)
char *string;
{
	static   char *buff;

	buff = string + strlen(string);
	while (*buff != '/' && buff > string) buff--;
	return( (char *) (*buff == '/'? ++buff : buff));
}
